/*
 * The gocommon package collects common packages to interact with the Joyent Public Cloud and Joyent Manta services.
 *
 * The gocommon package is structured as follow:
 *
 * 	- gocommon/client. Client for sending requests.
 * 	- gocommon/errors. Joyent specific errors.
 * 	- gocommon/http. HTTP client for sending requests.
 * 	- gocommon/jpc. This package provides common structures and functions across packages.
 * 	- gocommon/testing. Testing Suite for local testing.
 *
 * Copyright (c) 2016 Joyent Inc.
 * Written by Daniele Stroppa <daniele.stroppa@joyent.com>
 *
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

package gocommon
